# Generated by Django 2.0 on 2020-02-04 08:42

from django.db import migrations, models
import django.db.models.manager


class Migration(migrations.Migration):

    dependencies = [
        ('django_dramatiq', '0002_auto_20191104_1354'),
    ]

    operations = [
        migrations.AlterModelManagers(
            name='task',
            managers=[
                ('tasks', django.db.models.manager.Manager()),
            ],
        ),
        migrations.AlterField(
            model_name='task',
            name='status',
            field=models.CharField(choices=[('enqueued', 'Enqueued'), ('delayed', 'Delayed'), ('running', 'Running'), ('failed', 'Failed'), ('done', 'Done'), ('skipped', 'Skipped')], default='enqueued', max_length=8),
        ),
    ]
